from sempy_labs._helper_functions import (
    resolve_item_id,
    resolve_item_name_and_id,
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    delete_item,
    _decode_b64,
    create_item,
    resolve_workspace_name_and_id,
)
import pandas as pd
from typing import Any, Optional, List, Union, Literal
from uuid import UUID
from sempy._utils._log import log
import json
import sempy_labs._icons as icons
from os import PathLike


@log
def get_variable_library(
    variable_library: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns properties of the specified variable library.

    This is a wrapper function for the following API: `Items - Get Variable Library <https://learn.microsoft.com/rest/api/fabric/variablelibrary/items/get-variable-library>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    variable_library : str | uuid.UUID
        Name or ID of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the properties of the variable library.
    """

    columns = {
        "Variable Library Name": "string",
        "Variable Library Id": "string",
        "Description": "string",
        "Active Value Set Name": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    variable_library_id = resolve_item_id(
        item=variable_library, type="VariableLibrary", workspace=workspace
    )

    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/variableLibraries/{variable_library_id}",
        client="fabric_sp",
    )

    result = response.json()
    prop = result.get("properties", {})

    if prop:
        df = pd.DataFrame(
            [
                {
                    "Variable Library Name": result.get("displayName"),
                    "Variable Library Id": result.get("id"),
                    "Description": result.get("description"),
                    "Active Value Set Name": prop.get("activeValueSetName"),
                }
            ],
            columns=list(columns.keys()),
        )

        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_variable_libraries(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the variable libraries within a workspace.

    This is a wrapper function for the following API: `Items - List Variable Libraries <https://learn.microsoft.com/rest/api/fabric/variablelibrary/items/list-variable-libraries>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the variable libraries within a workspace.
    """

    columns = {
        "Variable Library Name": "string",
        "Variable Library Id": "string",
        "Description": "string",
        "Active Value Set Name": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/VariableLibraries",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})

            rows.append(
                {
                    "Variable Library Name": v.get("displayName"),
                    "Variable Library Id": v.get("id"),
                    "Description": v.get("description"),
                    "Active Value Set Name": prop.get("activeValueSetName"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def delete_variable_library(
    variable_library: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes a variable library.

    This is a wrapper function for the following API: `Items - Delete Variable Library https://learn.microsoft.com/rest/api/fabric/variablelibrary/items/delete-variable-library>`_.

    Parameters
    ----------
    navariable_libraryme: str | uuid.UUID
        Name or ID of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=variable_library, type="VariableLibrary", workspace=workspace)


@log
def get_variable_library_definition(
    variable_library: str | UUID,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
    return_dataframe: bool = False,
) -> dict | pd.DataFrame:
    """
    Gets the definition of a variable library.

    This is a wrapper function for the following API: `Items - Get Variable Library Definition <https://learn.microsoft.com/rest/api/fabric/variablelibrary/items/delete-variable-library>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    dict | pandas.DataFrame
        A dictionary showing the definition or a pandas dataframe showing the definition.
    """

    workspace_id = resolve_workspace_id(workspace)
    variable_library_id = resolve_item_id(
        item=variable_library, type="VariableLibrary", workspace=workspace
    )

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/variableLibraries/{variable_library_id}/getDefinition",
        method="post",
        client="fabric_sp",
        status_codes=None,
        lro_return_json=True,
    )

    if decode:
        definition = {"definition": {"parts": []}}

        for part in result.get("definition", {}).get("parts", []):
            path = part.get("path")
            payload = _decode_b64(part.get("payload"))
            payload_type = part.get("payloadType")
            definition["definition"]["parts"].append(
                {"path": path, "payload": payload, "payloadType": payload_type}
            )
    else:
        definition = result.copy()

    if return_dataframe:
        df = pd.DataFrame(definition["definition"]["parts"])
        df.columns = ["Path", "Payload", "Payload Type"]
        return df
    else:
        return definition


@log
def list_variables(
    variable_library: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Lists the variables in a variable library.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    variable_library : str | uuid.UUID
        Name or ID of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the variables within a variable library.
    """

    result = get_variable_library_definition(
        variable_library=variable_library,
        workspace=workspace,
        decode=True,
        return_dataframe=False,
    )

    columns = {
        "Variable Name": "string",
        "Note": "string",
        "Type": "string",
        "Value": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for part in result.get("definition").get("parts"):
        path = part.get("path")
        payload = json.loads(part.get("payload"))
        if path == "variables.json":

            for variable in payload.get("variables", []):
                rows.append(
                    {
                        "Variable Name": variable.get("name"),
                        "Note": variable.get("note"),
                        "Type": variable.get("type"),
                        "Value": variable.get("value"),
                    }
                )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    for part in result.get("definition", {}).get("parts", []):
        path = part.get("path")
        if path.startswith("valueSets") and path.endswith(".json"):
            payload = json.loads(part.get("payload"))
            value_set_name = payload.get("name")

            # Initialize the new column with None (or pd.NA)
            df[value_set_name] = None

            for override in payload.get("variableOverrides", []):
                variable_name = override.get("name")
                variable_value = override.get("value")

                # Set the value in the appropriate row and column
                df.loc[df["Variable Name"] == variable_name, value_set_name] = (
                    variable_value
                )

    return df


@log
def get_variable_values(
    variable_names: List[str],
    variable_library: Union[str, UUID],
    workspace: Optional[Union[str, UUID]] = None,
    value_set: Optional[str] = None,
) -> dict:
    """
    Gets the values of multiple variables from a variable library with a single call to list_variables.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    variable_names : List[str]
        A list of variable names to retrieve.
    variable_library : str | uuid.UUID
        Name or ID of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    value_set : str, default=None
        The name of the value set to use for variable overrides.
        If None, the active value set of the variable library will be used.

    Returns
    -------
    dict
        Dictionary mapping variable names to their corresponding values.
    """

    if isinstance(variable_names, str):
        variable_names = [variable_names]

    if value_set is None:
        vl_df = get_variable_library(
            variable_library=variable_library, workspace=workspace
        )
        if vl_df.empty:
            raise ValueError(
                f"{icons.red_dot} The variable library '{variable_library}' does not exist within the '{workspace}' workspace."
            )
        value_set = vl_df["Active Value Set Name"].iloc[0]

    df = list_variables(variable_library=variable_library, workspace=workspace)
    found_variables = df[df["Variable Name"].isin(variable_names)]

    missing = set(variable_names) - set(found_variables["Variable Name"])
    if missing:
        raise ValueError(
            f"{icons.red_dot} The following variables do not exist in the '{variable_library}' variable library: {', '.join(missing)}"
        )

    if value_set == "Default value set":
        value_set = "Value"
    if value_set not in df.columns:
        raise ValueError(
            f"{icons.red_dot} The value set '{value_set}' does not exist in the variable library '{variable_library}' within the '{workspace}' workspace."
        )

    return dict(zip(found_variables["Variable Name"], found_variables[value_set]))


@log
def get_variable_value(
    variable_name: str,
    variable_library: str | UUID,
    workspace: Optional[str | UUID] = None,
    value_set: Optional[str] = None,
) -> Any:
    """
    Gets the value of a single variable in a variable library.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    variable_name : str
        Name of the variable.
    variable_library : str | uuid.UUID
        Name or ID of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    value_set : str, default=None
        The name of the value set to use for variable overrides.
        If None, the active value set of the variable library will be used.

    Returns
    -------
    Any
        The value of the variable.
    """

    return get_variable_values(
        variable_names=[variable_name],
        variable_library=variable_library,
        workspace=workspace,
        value_set=value_set,
    )[variable_name]


def _encode_b64(content: Any) -> str:
    import base64

    json_content = json.dumps(content, separators=(",", ":"))

    return base64.b64encode(json_content.encode("utf-8")).decode("utf-8")

    # if isinstance(content, (dict, list)):
    #    content = json.dumps(content, ensure_ascii=False)

    # if isinstance(content, str):
    #    content = content.encode("utf-8")

    # return base64.b64encode(content).decode("utf-8")


@log
def create_variable_library(
    name: str,
    variables: List[dict],
    value_sets: List[dict],
    value_sets_order: Optional[List[str]] = None,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    folder: Optional[str | PathLike] = None,
):
    """
    Creates a variable library.

    This is a wrapper function for the following API: `Items - Create Variable Library <https://learn.microsoft.com/rest/api/fabric/variablelibrary/items/create-variable-library>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        Name of the variable library.
    variables : List[dict]
        A list of variable definitions. Each variable definition should be a dictionary with keys such as 'name', 'type', 'value', and 'note'.

        Example:

        variables = [
            {
            "name": "variable1",
            "note": "Some optional note",
            "type": "String",
            "value": "Some string value"
            },
            {
            "name": "variableWithoutNote",
            "type": "Boolean",
            "value": True
            }
        ]
    value_sets : List[dict]
        A list of value set definitions. Each value set definition should be a dictionary with keys such as 'name' and 'variableOverrides'.

        Example:

        value_sets = [
            {
                "name": "valueSet1",
                "variableOverrides": [
                    {
                        "name": "variable1",
                        "value": "Overridden string value"
                    },
                    {
                        "name": "variableWithoutNote",
                        "value": False
                    }
                ]
            },            {
                "name": "valueSet0",
                "variableOverrides": [
                    {
                        "name": "variable1",
                        "value": "Another overridden string value"
                    },
                    {
                        "name": "variableWithoutNote",
                        "value": True
                    }
                ]
            }
        ]

    value_sets_order : List[str], default=None
        The order of value sets. If None, a default order will be used.
    description : str, default=None
        Description of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    folder : str | PathLike, default=None
        The folder path within the workspace to create the variable library in.
    """

    schema_start = "https://developer.microsoft.com/json-schemas/fabric/item/variableLibrary/definition"
    schema_end = "1.0.0/schema.json"

    # Variable type normalization and validation
    variable_types = {
        "boolean": "Boolean",
        "datetime": "DateTime",
        "number": "Number",
        "integer": "Integer",
        "string": "String",
    }

    for v in variables:
        raw_type = v.get("type")

        if not isinstance(raw_type, str):
            raise ValueError(
                f"{icons.red_dot} Variable '{v.get('name')}' has an invalid type: {raw_type}"
            )

        normalized_key = raw_type.strip().lower()

        if normalized_key not in variable_types:
            raise ValueError(
                f"{icons.red_dot} Invalid variable type '{raw_type}' for variable '{v.get('name')}'. "
                f"Valid types are: {', '.join(variable_types.values())}"
            )

        v["type"] = variable_types[normalized_key]

    variables_content = {
        "$schema": f"{schema_start}/variables/{schema_end}",
        "variables": variables,
    }

    definition = {
        "format": "VariableLibraryV1",
        "parts": [
            {
                "path": "variables.json",
                "payload": _encode_b64(variables_content),
                "payloadType": "InlineBase64",
            },
        ],
    }

    # Value Sets
    if value_sets:
        for vs in value_sets:
            vs_name = vs.get("name")
            vs_content = {
                "$schema": f"{schema_start}/valueSet/{schema_end}",
                "name": vs_name,
                "variableOverrides": vs.get("variableOverrides", []),
            }
            definition["parts"].append(
                {
                    "path": f"valueSets/{vs_name}.json",
                    "payload": _encode_b64(vs_content),
                    "payloadType": "InlineBase64",
                }
            )

    # Settings
    if value_sets_order:
        settings_content = {
            "$schema": f"{schema_start}/settings/{schema_end}",
            "valueSetsOrder": value_sets_order,
        }
    else:
        settings_content = {
            "$schema": f"{schema_start}/settings/{schema_end}",
            "valueSetsOrder": [vs.get("name") for vs in value_sets],
        }

    definition["parts"].append(
        {
            "path": "settings.json",
            "payload": _encode_b64(settings_content),
            "payloadType": "InlineBase64",
        }
    )
    # return definition
    create_item(
        name=name,
        type="VariableLibrary",
        description=description,
        definition=definition,
        workspace=workspace,
        folder=folder,
    )


@log
def update_variable_library(
    variable_library: str | UUID,
    name: Optional[str] = None,
    description: Optional[str] = None,
    default_value_set: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the properties of the specified variable library.

    This is a wrapper function for the following API: `Items - Update Variable Library <https://learn.microsoft.com/rest/api/fabric/variablelibrary/items/update-variable-library>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    variable_library : str | uuid.UUID
        Name or ID of the variable library.
    name : str, default=None
        New name of the variable library.
    description : str, default=None
        New description of the variable library.
    default_value_set : str, default=None
        New default value set name of the variable library.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=variable_library, type="VariableLibrary", workspace=workspace_id
    )

    payload = {}

    if name:
        payload["displayName"] = name
    if description:
        payload["description"] = description
    if default_value_set:
        payload["properties"] = {"activeValueSetName": default_value_set}

    if not payload:
        print(
            f"{icons.info} No updates provided for the '{item_name}' variable library within the '{workspace_name}' workspace."
        )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/VariableLibraries/{item_id}",
        method="patch",
        payload=payload,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{item_name}' variable library within the '{workspace_name}' workspace has been updated."
    )


@log
def update_variable(
    variable_library: str | UUID,
    name: str,
    new_name: Optional[str] = None,
    type: Literal["Boolean", "DateTime", "Number", "Integer", "String"] = None,
    value: Optional[str] = None,
    note: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the properties of the specified variable within a variable library.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    variable_library : str | uuid.UUID
        Name or ID of the variable library.
    name : str
        Name of the variable.
    new_name : str, default=None
        New name of the variable. Defaults to None which keeps the same name.
    type : typing.Literal["Boolean", "DateTime", "Number", "Integer", "String"], default=None
        New type of the variable. Valid types are: "Boolean", "DateTime", "Number", "Integer", "String". Defaults to None which keeps the same type.
    value : str, default=None
        New value of the variable. Defaults to None which keeps the same value.
    note : str, default=None
        New note of the variable. Defaults to None which keeps the same note.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=variable_library, type="VariableLibrary", workspace=workspace_id
    )

    definition = get_variable_library_definition(
        variable_library=item_id,
        workspace=workspace_id,
        decode=True,
        return_dataframe=False,
    )

    if new_name is None and type is None and value is None and note is None:
        print(
            f"{icons.info} No updates provided for the variable '{name}' in the '{item_name}' variable library within the '{workspace_name}' workspace."
        )
        return

    for part in definition.get("definition", {}).get("parts", []):
        payload = json.loads(part.get("payload"))
        if part.get("path") != "variables.json":
            part["payload"] = _encode_b64(payload)
        else:
            variables = payload.get("variables", [])
            if name not in [v.get("name") for v in variables]:
                raise ValueError(
                    f"{icons.red_dot} The variable '{name}' does not exist in the '{item_name}' variable library within the '{workspace_name}' workspace."
                )

            for variable in variables:
                if variable.get("name") == name:
                    if new_name is not None:
                        variable["name"] = new_name
                    if type is not None:
                        variable["type"] = type
                    if note is not None:
                        variable["note"] = note
                    if value is not None:
                        variable["value"] = value
                    break
            part["payload"] = _encode_b64(payload)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/variableLibraries/{item_id}/updateDefinition",
        method="post",
        payload=definition,
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} The variable '{name}' in the '{item_name}' variable library within the '{workspace_name}' workspace has been updated."
    )



import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs.mounted_data_factory as mdf


@log
def list_mounted_data_factories(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Shows a list of mounted data factories from the specified workspace.

    This is a wrapper function for the following API: `Items - List Mounted Data Factories <https://learn.microsoft.com/rest/api/fabric/mounteddatafactory/items/list-mounted-data-factories>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of mounted data factories from the specified workspace.
    """

    return mdf.list_mounted_data_factories(workspace=workspace)


@log
def get_mounted_data_factory_definition(
    mounted_data_factory: str | UUID, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Returns the specified MountedDataFactory public definition.

    This is a wrapper function for the following API: `Items - Get Mounted Data Factory Definition <https://learn.microsoft.com/rest/api/fabric/mounteddatafactory/items/get-mounted-data-factory-definition>`_.

    Parameters
    ----------
    mounted_data_factory : str | uuid.UUID
        The name or ID of the mounted data factory.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    dict
        The 'mountedDataFactory-content.json' file from the mounted data factory definition.
    """

    return mdf.get_mounted_data_factory_definition(
        mounted_data_factory=mounted_data_factory, workspace=workspace
    )


@log
def delete_mounted_data_factory(
    mounted_data_factory: str | UUID, workspace: Optional[str | UUID]
):
    """
    Deletes the specified mounted data factory.

    This is a wrapper function for the following API: `Items - Delete Mounted Data Factory <https://learn.microsoft.com/rest/api/fabric/mounteddatafactory/items/delete-mounted-data-factory>`_.

    Parameters
    ----------
    mounted_data_factory : str | uuid.UUID
        The name or ID of the mounted data factory.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    mdf.delete_mounted_data_factory(
        mounted_data_factory=mounted_data_factory, workspace=workspace
    )



from typing import Optional
from uuid import UUID
import pandas as pd
from sempy_labs._helper_functions import (
    _create_dataframe,
    _base_api,
    resolve_workspace_id,
    _update_dataframe_datatypes,
)
from sempy._utils._log import log


@log
def list_dashboards(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows a list of the dashboards within a workspace.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the dashboards within a workspace.
    """

    columns = {
        "Dashboard ID": "string",
        "Dashboard Name": "string",
        "Read Only": "bool",
        "Web URL": "string",
        "Embed URL": "string",
        "Data Classification": "string",
        "Users": "string",
        "Subscriptions": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/dashboards", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Dashboard ID": v.get("id"),
                "Dashboard Name": v.get("displayName"),
                "Read Only": v.get("isReadOnly"),
                "Web URL": v.get("webUrl"),
                "Embed URL": v.get("embedUrl"),
                "Data Classification": v.get("dataClassification"),
                "Users": v.get("users"),
                "Subscriptions": v.get("subscriptions"),
            }
        )
    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



from ._functions import (
    get_variable_library,
    list_variables,
    list_variable_libraries,
    delete_variable_library,
    get_variable_library_definition,
    get_variable_values,
    get_variable_value,
    create_variable_library,
    update_variable_library,
    update_variable,
)

__all__ = [
    "get_variable_library",
    "list_variables",
    "list_variable_libraries",
    "delete_variable_library",
    "get_variable_library_definition",
    "get_variable_values",
    "get_variable_value",
    "create_variable_library",
    "update_variable_library",
    "update_variable",
]



from typing import Literal, Optional, List
import sempy_labs._icons as icons
import pandas as pd
from uuid import UUID
from sempy_labs.admin._basic_functions import list_workspaces
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _is_valid_uuid,
)
from sempy._utils._log import log


@log
def resolve_domain_id(domain: Optional[str | UUID] = None, **kwargs) -> UUID:
    """
    Obtains the domain Id for a given domain name.

    Parameters
    ----------
    domain_name : str | uuid.UUID
        The domain name or ID

    Returns
    -------
    uuid.UUID
        The domain Id.
    """

    if "domain_name" in kwargs:
        domain = kwargs["domain_name"]
        print(
            f"{icons.warning} The 'domain_name' parameter is deprecated. Please use 'domain' instead."
        )

    if domain is None:
        raise ValueError(f"{icons.red_dot} Please provide a domain.")

    if _is_valid_uuid(domain):
        return domain

    dfL = list_domains()
    dfL_filt = dfL[dfL["Domain Name"] == domain]
    if dfL_filt.empty:
        raise ValueError(f"{icons.red_dot} '{domain}' is not a valid domain name.")

    return dfL_filt["Domain ID"].iloc[0]


@log
def resolve_domain_name(domain: Optional[str | UUID], **kwargs) -> UUID:
    """
    Obtains the domain name for a given domain ID.

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID

    Returns
    -------
    str
        The domain Name.
    """

    if "domain_name" in kwargs:
        domain = kwargs["domain_name"]
        print(
            f"{icons.warning} The 'domain_name' parameter is deprecated. Please use 'domain' instead."
        )

    if domain is None:
        raise ValueError(f"{icons.red_dot} Please provide a domain.")

    if not _is_valid_uuid(domain):
        return domain

    dfL = list_domains()
    dfL_filt = dfL[dfL["Domain ID"] == domain]
    if dfL_filt.empty:
        raise ValueError(f"{icons.red_dot} '{domain}' is not a valid domain name.")

    return dfL_filt["Domain Name"].iloc[0]


@log
def list_domains(non_empty_only: bool = False) -> pd.DataFrame:
    """
    Shows a list of domains.

    This is a wrapper function for the following API: `Domains - List Domains <https://learn.microsoft.com/rest/api/fabric/admin/domains/list-domains>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    non_empty_only : bool, default=False
        When True, only return domains that have at least one workspace containing an item.
        Defaults to False.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the domains.
    """

    columns = {
        "Domain ID": "string",
        "Domain Name": "string",
        "Description": "string",
        "Parent Domain ID": "string",
        "Contributors Scope": "string",
    }
    df = _create_dataframe(columns=columns)

    url = "/v1/admin/domains"
    if non_empty_only:
        url = f"{url}?nonEmptyOnly=True"

    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("domains", []):
        rows.append(
            {
                "Domain ID": v.get("id"),
                "Domain Name": v.get("displayName"),
                "Description": v.get("description"),
                "Parent Domain ID": v.get("parentDomainId"),
                "Contributors Scope": v.get("contributorsScope"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_domain_workspaces(domain: Optional[str] = None, **kwargs) -> pd.DataFrame:
    """
    Shows a list of workspaces within the domain.

    This is a wrapper function for the following API: `Domains - List Domain Workspaces <https://learn.microsoft.com/rest/api/fabric/admin/domains/list-domain-workspaces>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of workspaces within the domain.
    """

    if "domain_name" in kwargs:
        domain = kwargs["domain_name"]
        print(
            f"{icons.warning} The 'domain_name' parameter is deprecated. Please use 'domain' instead."
        )

    if domain is None:
        raise ValueError(f"{icons.red_dot} Please provide a domain.")

    domain_id = resolve_domain_id(domain)

    columns = {
        "Workspace ID": "string",
        "Workspace Name": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1/admin/domains/{domain_id}/workspaces", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Workspace ID": v.get("id"),
                "Workspace Name": v.get("displayName"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def create_domain(
    domain_name: str,
    description: Optional[str] = None,
    parent_domain: Optional[str | UUID] = None,
    **kwargs,
):
    """
    Creates a new domain.

    This is a wrapper function for the following API: `Domains - Create Domain <https://learn.microsoft.com/rest/api/fabric/admin/domains/create-domain>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain_name : str
        The domain name.
    description : str, default=None
        The domain description.
    parent_domain : str | uuid.UUID, default=None
        The parent domain name or ID.
    """

    if "parent_domain_name" in kwargs:
        parent_domain = kwargs["parent_domain_name"]
        print(
            f"{icons.warning} The 'parent_domain_name' parameter is deprecated. Please use 'parent_domain' instead."
        )

    if parent_domain is not None:
        parent_domain_id = resolve_domain_id(domain=parent_domain)

    payload = {}
    payload["displayName"] = domain_name
    if description is not None:
        payload["description"] = description
    if parent_domain is not None:
        payload["parentDomainId"] = parent_domain_id

    _base_api(
        request="/v1/admin/domains",
        method="post",
        payload=payload,
        status_codes=201,
        client="fabric_sp",
    )

    print(f"{icons.green_dot} The '{domain_name}' domain has been created.")


@log
def delete_domain(domain: Optional[str | UUID], **kwargs):
    """
    Deletes a domain.

    This is a wrapper function for the following API: `Domains - Delete Domain <https://learn.microsoft.com/rest/api/fabric/admin/domains/delete-domain>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    """

    if "domain_name" in kwargs:
        domain = kwargs["domain_name"]
        print(
            f"{icons.warning} The 'domain_name' parameter is deprecated. Please use 'domain' instead."
        )

    if domain is None:
        raise ValueError(f"{icons.red_dot} Please provide a domain.")

    domain_id = resolve_domain_id(domain)
    _base_api(
        request=f"/v1/admin/domains/{domain_id}", method="delete", client="fabric_sp"
    )

    print(f"{icons.green_dot} The '{domain}' domain has been deleted.")


@log
def update_domain(
    domain: Optional[str | UUID] = None,
    description: Optional[str] = None,
    contributors_scope: Optional[str] = None,
    **kwargs,
):
    """
    Updates a domain's properties.

    This is a wrapper function for the following API: `Domains - Update Domain <https://learn.microsoft.com/rest/api/fabric/admin/domains/update-domain>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    description : str, default=None
        The domain description.
    contributors_scope : str, default=None
        The domain `contributor scope <https://learn.microsoft.com/rest/api/fabric/admin/domains/update-domain?tabs=HTTP#contributorsscopetype>`_.
    """

    if "domain_name" in kwargs:
        domain = kwargs["domain_name"]
        print(
            f"{icons.warning} The 'domain_name' parameter is deprecated. Please use 'domain' instead."
        )

    if domain is None:
        raise ValueError(f"{icons.red_dot} Please provide a domain.")
    contributors_scopes = ["AdminsOnly", "AllTenant", "SpecificUsersAndGroups"]

    if contributors_scope not in contributors_scopes:
        raise ValueError(
            f"{icons.red_dot} Invalid contributors scope. Valid options: {contributors_scopes}."
        )

    domain_id = resolve_domain_id(domain)
    domain_name = resolve_domain_name(domain)

    payload = {}
    payload["displayName"] = domain_name

    if description is not None:
        payload["description"] = description
    if contributors_scope is not None:
        payload["contributorsScope"] = contributors_scope

    _base_api(
        request=f"/v1/admin/domains/{domain_id}",
        method="patch",
        payload=payload,
        client="fabric_sp",
    )

    print(f"{icons.green_dot} The '{domain_name}' domain has been updated.")


@log
def assign_domain_workspaces_by_capacities(
    domain: str | UUID,
    capacity_names: str | List[str],
    **kwargs,
):
    """
    Assigns all workspaces that reside on the specified capacities to the specified domain.

    This is a wrapper function for the following API: `Domains - Assign Domain Workspaces By Capacities <https://learn.microsoft.com/rest/api/fabric/admin/domains/assign-domain-workspaces-by-capacities>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    capacity_names : str | List[str]
        The capacity names.
    """

    from sempy_labs.admin._capacities import list_capacities

    if "domain_name" in kwargs:
        domain = kwargs["domain_name"]
        print(
            f"{icons.warning} The 'domain_name' parameter is deprecated. Please use 'domain' instead."
        )

    if domain is None:
        raise ValueError(f"{icons.red_dot} Please provide a domain.")
    domain_id = resolve_domain_id(domain)

    if isinstance(capacity_names, str):
        capacity_names = [capacity_names]

    dfC = list_capacities()

    # Check for invalid capacities
    invalid_capacities = [
        name for name in capacity_names if name not in dfC["Display Name"].values
    ]

    if len(invalid_capacities) == 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_capacities} capacity is not valid."
        )
    elif len(invalid_capacities) > 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_capacities} capacities are not valid."
        )

    # Get list of capacity Ids for the payload
    dfC_filt = dfC[dfC["Display Name"].isin(capacity_names)]
    capacity_list = list(dfC_filt["Id"].str.upper())

    payload = {"capacitiesIds": capacity_list}

    _base_api(
        request=f"/v1/admin/domains/{domain_id}/assignWorkspacesByCapacities",
        method="post",
        payload=payload,
        lro_return_status_code=True,
        status_codes=202,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The workspaces in the {capacity_names} capacities have been assigned to the '{domain}' domain."
    )


@log
def assign_domain_workspaces(
    domain: str | UUID, workspace: str | UUID | List[str | UUID], **kwargs
):
    """
    Assigns workspaces to the specified domain by workspace.

    This is a wrapper function for the following API: `Domains - Assign Domain Workspaces By Ids <https://learn.microsoft.com/rest/api/fabric/admin/domains/assign-domain-workspaces-by-ids>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    workspace : str | uuid.UUID | List[str | uuid.UUID]
        The Fabric workspace name or IDs.
    """

    domain_id = resolve_domain_id(domain)

    if "workspace_names" in kwargs:
        workspace = kwargs["workspace_names"]
        print(
            f"{icons.warning} The 'workspace_names' parameter is deprecated. Please use 'workspace' instead."
        )

    if isinstance(workspace, str):
        workspace = [workspace]
    workspace_list = []

    for w in workspace:
        if _is_valid_uuid(w):
            workspace_list.append(w)
        else:
            dfW = list_workspaces(workspace=w)
            if not dfW.empty:
                workspace_list.append(dfW["Id"].iloc[0])

    # Check for invalid capacities
    # invalid_workspaces = [
    #    name for name in workspace_names if name not in dfW["Name"].values
    # ]

    # if len(invalid_workspaces) == 1:
    #    raise ValueError(
    #        f"{icons.red_dot} The {invalid_workspaces} workspace is not valid."
    #    )
    # elif len(invalid_workspaces) > 1:
    #    raise ValueError(
    #        f"{icons.red_dot} The {invalid_workspaces} workspaces are not valid."
    #    )

    # dfW_filt = dfW[dfW["Name"].isin(workspace_names)]
    # workspace_list = list(dfW_filt["Id"])

    payload = {"workspacesIds": workspace_list}

    _base_api(
        request=f"/v1/admin/domains/{domain_id}/assignWorkspaces",
        method="post",
        payload=payload,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The {workspace} workspaces have been assigned to the '{domain}' domain."
    )


@log
def unassign_all_domain_workspaces(domain: str | UUID):
    """
    Unassigns all workspaces from the specified domain.

    This is a wrapper function for the following API: `Domains - Unassign All Domain Workspaces <https://learn.microsoft.com/rest/api/fabric/admin/domains/unassign-all-domain-workspaces>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    """

    domain_id = resolve_domain_id(domain)

    _base_api(
        request=f"/v1/admin/domains/{domain_id}/unassignAllWorkspaces",
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} All workspaces assigned to the '{domain}' domain have been unassigned."
    )


@log
def unassign_domain_workspaces(
    domain: str | UUID,
    workspace_names: str | List[str],
):
    """
    Unassigns workspaces from the specified domain by workspace.

    This is a wrapper function for the following API: `Domains - Unassign Domain Workspaces By Ids <https://learn.microsoft.com/rest/api/fabric/admin/domains/unassign-domain-workspaces-by-ids>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    workspace_names : str | List[str]
        The Fabric workspace(s).
    """

    domain_id = resolve_domain_id(domain)

    if isinstance(workspace_names, str):
        workspace_names = [workspace_names]

    dfW = list_workspaces()

    # Check for invalid capacities
    invalid_workspaces = [
        name for name in workspace_names if name not in dfW["Name"].values
    ]

    if len(invalid_workspaces) == 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_workspaces} workspace is not valid."
        )
    elif len(invalid_workspaces) > 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_workspaces} workspaces are not valid."
        )

    dfW_filt = dfW[dfW["Name"].isin(workspace_names)]
    workspace_list = list(dfW_filt["Id"])

    payload = {"workspacesIds": workspace_list}

    _base_api(
        request=f"/v1/admin/domains/{domain_id}/unassignWorkspaces",
        method="post",
        payload=payload,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The {workspace_names} workspaces assigned to the '{domain}' domain have been unassigned."
    )


@log
def sync_role_assignments_to_subdomains(
    domain: str | UUID, role: Literal["Admin", "Contributor"]
):
    """
    Sync the role assignments from the specified domain to its subdomains.

    This is a wrapper function for the following API: `Domains - Sync Role Assignments To Subdomains <https://learn.microsoft.com/rest/api/fabric/admin/domains/sync-role-assignments-to-subdomains>`_.

    Parameters
    ----------
    domain : str | uuid.UUID
        The domain name or ID.
    role : typing.Literal["Admin", "Contributor"]
        The role to sync. Valid options: 'Admin', 'Contributor'.
    """

    role = role.capitalize()
    valid_roles = ["Admin", "Contributor"]
    if role not in valid_roles:
        raise ValueError(f"{icons.red_dot} Invalid role. Valid options: {valid_roles}.")

    domain_id = resolve_domain_id(domain)

    payload = {"role": role}

    _base_api(
        request=f"/v1/admin/domains/{domain_id}/roleAssignments/syncToSubdomains",
        method="post",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{role}' role assignments for the '{domain}' domain have been synced to its subdomains."
    )



from ._items import (
    create_external_data_share,
    delete_external_data_share,
    list_external_data_shares_in_item,
    revoke_external_data_share,
)

__all__ = [
    "create_external_data_share",
    "delete_external_data_share",
    "list_external_data_shares_in_item",
    "revoke_external_data_share",
]



import sempy.fabric as fabric
import pandas as pd
from typing import Optional
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    resolve_workspace_id,
)
import sempy_labs._icons as icons
from uuid import UUID


@log
def export_model_to_onelake(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    destination_lakehouse: Optional[str] = None,
    destination_workspace: Optional[str] = None,
):
    """
    Exports a semantic model's tables to delta tables in the lakehouse. Creates shortcuts to the tables if a lakehouse is specified.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    destination_lakehouse : str, default=None
        The name of the Fabric lakehouse where shortcuts will be created to access the delta tables created by the export. If the lakehouse specified does not exist, one will be created with that name. If no lakehouse is specified, shortcuts will not be created.
    destination_workspace : str, default=None
        The name of the Fabric workspace in which the lakehouse resides.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    if destination_workspace is None:
        destination_workspace = workspace_name
        destination_workspace_id = workspace_id
    else:
        destination_workspace_id = resolve_workspace_id(workspace=destination_workspace)

    tmsl = f"""
    {{
    'export': {{
    'layout': 'delta',
    'type': 'full',
    'objects': [
        {{
        'database': '{dataset_name}'
        }}
    ]
    }}
    }}
    """

    # Export model's tables as delta tables
    try:
        fabric.execute_tmsl(script=tmsl, workspace=workspace_id)
        print(
            f"{icons.green_dot} The '{dataset_name}' semantic model's tables have been exported as delta tables to the '{workspace_name}' workspace.\n"
        )
    except Exception as e:
        raise ValueError(
            f"{icons.red_dot} The '{dataset_name}' semantic model's tables have not been exported as delta tables to the '{workspace_name}' workspace.\nMake sure you enable OneLake integration for the '{dataset_name}' semantic model. Follow the instructions here: https://learn.microsoft.com/power-bi/enterprise/onelake-integration-overview#enable-onelake-integration"
        ) from e

    # Create shortcuts if destination lakehouse is specified
    if destination_lakehouse is not None:
        # Destination...
        dfI_Dest = fabric.list_items(workspace=destination_workspace, type="Lakehouse")
        dfI_filt = dfI_Dest[(dfI_Dest["Display Name"] == destination_lakehouse)]

        if len(dfI_filt) == 0:
            print(
                f"{icons.red_dot} The '{destination_lakehouse}' lakehouse does not exist within the '{destination_workspace}' workspace."
            )
            # Create lakehouse
            destination_lakehouse_id = fabric.create_lakehouse(
                display_name=destination_lakehouse, workspace=destination_workspace
            )
            print(
                f"{icons.green_dot} The '{destination_lakehouse}' lakehouse has been created within the '{destination_workspace}' workspace.\n"
            )
        else:
            destination_lakehouse_id = dfI_filt["Id"].iloc[0]

        # Source...
        dfI_Source = fabric.list_items(workspace=workspace_id, type="SemanticModel")
        dfI_filtSource = dfI_Source[(dfI_Source["Display Name"] == dataset)]
        sourceLakehouseId = dfI_filtSource["Id"].iloc[0]

        # Valid tables
        dfP = fabric.list_partitions(
            dataset=dataset_id,
            workspace=workspace_id,
            additional_xmla_properties=["Parent.SystemManaged"],
        )
        dfP_filt = dfP[
            (dfP["Mode"] == "Import")
            & (dfP["Source Type"] != "CalculationGroup")
            & (dfP["Parent System Managed"] == False)
        ]
        dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id)
        tmc = pd.DataFrame(dfP.groupby("Table Name")["Mode"].nunique()).reset_index()
        oneMode = tmc[tmc["Mode"] == 1]
        tableAll = dfP_filt[
            dfP_filt["Table Name"].isin(dfC["Table Name"].values)
            & (dfP_filt["Table Name"].isin(oneMode["Table Name"].values))
        ]
        tables = tableAll["Table Name"].unique()

        client = fabric.FabricRestClient()

        print(f"{icons.in_progress} Creating shortcuts...\n")
        for tableName in tables:
            tablePath = f"Tables/{tableName}"
            shortcutName = tableName.replace(" ", "")
            request_body = {
                "path": "Tables",
                "name": shortcutName,
                "target": {
                    "oneLake": {
                        "workspaceId": workspace_id,
                        "itemId": sourceLakehouseId,
                        "path": tablePath,
                    }
                },
            }

            try:
                response = client.post(
                    f"/v1/workspaces/{destination_workspace_id}/items/{destination_lakehouse_id}/shortcuts",
                    json=request_body,
                )
                if response.status_code == 201:
                    print(
                        f"{icons.bullet} The shortcut '{shortcutName}' was created in the '{destination_lakehouse}' lakehouse within the '{destination_workspace}' workspace. It is based on the '{tableName}' table in the '{dataset_name}' semantic model within the '{workspace_name}' workspace.\n"
                    )
                else:
                    print(response.status_code)
            except Exception as e:
                raise ValueError(
                    f"{icons.red_dot} Failed to create a shortcut for the '{tableName}' table."
                ) from e



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    _is_valid_uuid,
)
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def list_reports(
    top: Optional[int] = None,
    skip: Optional[int] = None,
    filter: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows a list of reports for the organization.

    This is a wrapper function for the following API: `Admin - Reports GetReportsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/reports-get-reports-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    top : int, default=None
        Returns only the first n results.
    skip : int, default=None
        Skips the first n results.
    filter : str, default=None
        Returns a subset of a results based on Odata filter query parameter condition.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of reports for the organization.
    """

    columns = {
        "Report Id": "string",
        "Report Name": "string",
        "Type": "string",
        "Web URL": "string",
        "Embed URL": "string",
        "Dataset Id": "string",
        "Created Date": "datetime_coerce",
        "Modified Date": "datetime_coerce",
        "Created By": "string",
        "Modified By": "string",
        "Sensitivity Label Id": "string",
        "Users": "string",
        "Subscriptions": "string",
        "Workspace Id": "string",
        "Report Flags": "int_fillna",
    }

    df = _create_dataframe(columns=columns)

    url = "/v1.0/myorg/admin/reports?"
    if top is not None:
        url += f"$top={top}&"
    if skip is not None:
        url += f"$skip={skip}&"
    if filter is not None:
        url += f"$filter={filter}&"

    url.rstrip("$").rstrip("?")
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Report Id": v.get("id"),
                "Report Name": v.get("name"),
                "Type": v.get("reportType"),
                "Web URL": v.get("webUrl"),
                "Embed URL": v.get("embedUrl"),
                "Dataset Id": v.get("datasetId"),
                "Created Date": v.get("createdDateTime"),
                "Modified Date": v.get("modifiedDateTime"),
                "Created By": v.get("createdBy"),
                "Modified By": v.get("modifiedBy"),
                "Sensitivity Label Id": v.get("sensitivityLabel", {}).get("labelId"),
                "Users": v.get("users"),
                "Subscriptions": v.get("subscriptions"),
                "Workspace Id": v.get("workspaceId"),
                "Report Flags": v.get("reportFlags"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def _resolve_report_id(report: str | UUID) -> str:
    if _is_valid_uuid(report):
        return report
    else:
        df = list_reports()
        df_filt = df[df["Report Name"] == report]
        if df_filt.empty:
            raise ValueError(f"{icons.red_dot} The '{report}' report does not exist.")
        return df_filt["Report Id"].iloc[0]


@log
def list_report_users(report: str | UUID) -> pd.DataFrame:
    """
    Shows a list of users that have access to the specified report.

    This is a wrapper function for the following API: `Admin - Reports GetDatasetUsersAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/reports-get-report-users-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    report : str | uuid.UUID
        The name or ID of the report.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users that have access to the specified report.
    """

    report_id = _resolve_report_id(report)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "Report User Access Right": "string",
        "Identifier": "string",
        "Graph Id": "string",
        "Principal Type": "string",
    }

    df = _create_dataframe(columns=columns)

    url = f"/v1.0/myorg/admin/reports/{report_id}/users"
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "User Name": v.get("displayName"),
                "Email Address": v.get("emailAddress"),
                "Report User Access Right": v.get("reportUserAccessRight"),
                "Identifier": v.get("identifier"),
                "Graph Id": v.get("graphId"),
                "Principal Type": v.get("principalType"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_report_subscriptions(report: str | UUID) -> pd.DataFrame:
    """
    Shows a list of report subscriptions along with subscriber details. This is a preview API call.

    This is a wrapper function for the following API: `Admin - Reports GetReportSubscriptionsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/reports-get-report-subscriptions-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    report : str | uuid.UUID
        The name or ID of the report.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of report subscriptions along with subscriber details. This is a preview API call.
    """

    report_id = _resolve_report_id(report)

    columns = {
        "Subscription Id": "string",
        "Title": "string",
        "Artifact Id": "string",
        "Artifact Name": "string",
        "Sub Artifact Name": "string",
        "Artifact Type": "string",
        "Is Enabled": "bool",
        "Frequency": "string",
        "Start Date": "datetime",
        "End Date": "string",
        "Link To Content": "bool",
        "Preview Image": "bool",
        "Attachment Format": "string",
        "Users": "list",
    }

    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1.0/myorg/admin/reports/{report_id}/subscriptions",
        client="fabric_sp",
    )

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Subscription Id": v.get("id"),
                "Title": v.get("title"),
                "Artifact Id": v.get("artifactId"),
                "Artifact Name": v.get("artifactDisplayName"),
                "Sub Artifact Name": v.get("subArtifactDisplayName"),
                "Artifact Type": v.get("artifactType"),
                "Is Enabled": v.get("isEnabled"),
                "Frequency": v.get("frequency"),
                "Start Date": v.get("startDate"),
                "End Date": v.get("endDate"),
                "Link To Content": v.get("linkToContent"),
                "Preview Image": v.get("previewImage"),
                "Attachment Format": v.get("attachmentFormat"),
                "Users": v.get("users", []),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



from ._items import (
    get_mounted_data_factory_definition,
    list_mounted_data_factories,
    delete_mounted_data_factory,
)

__all__ = [
    "list_mounted_data_factories",
    "get_mounted_data_factory_definition",
    "delete_mounted_data_factory",
]



from sempy_labs.lakehouse._blobs import _request_blob_api
from sempy_labs._helper_functions import (
    _xml_to_dict,
)
from datetime import datetime, timedelta, timezone
import xml.etree.ElementTree as ET
from sempy._utils._log import log


@log
def get_user_delegation_key():
    """
    Gets a key that can be used to sign a user delegation SAS (shared access signature). A user delegation SAS grants access to Azure Blob Storage resources by using Microsoft Entra credentials.

    This is a wrapper function for the following API: `Get User Delegation Key <https://learn.microsoft.com/rest/api/storageservices/get-user-delegation-key>`_.

    Returns
    -------
    str
        The user delegation key value.
    """

    utc_now = datetime.now(timezone.utc)
    start_time = utc_now + timedelta(minutes=2)
    expiry_time = start_time + timedelta(minutes=60)
    start_str = start_time.strftime("%Y-%m-%dT%H:%M:%SZ")
    expiry_str = expiry_time.strftime("%Y-%m-%dT%H:%M:%SZ")

    payload = f"""<?xml version="1.0" encoding="utf-8"?>
    <KeyInfo>
        <Start>{start_str}</Start>
        <Expiry>{expiry_str}</Expiry>
    </KeyInfo>"""

    response = _request_blob_api(
        request="?restype=service&comp=userdelegationkey",
        method="post",
        payload=payload,
    )

    root = ET.fromstring(response.content)
    response_json = _xml_to_dict(root)

    return response_json.get("UserDelegationKey", {}).get("Value", None)



import sempy
import requests
import zipfile
from pathlib import Path

lib_name = "semanticlinklabs"
lib_version = "0.13.0"

NUGET_BASE_URL = "https://www.nuget.org/api/v2/package"
current_dir = Path(__file__).parent
nuget_dir = current_dir / "nuget_dlls"


def find_lib_folder(pkg_folder: Path) -> Path:
    lib_base = pkg_folder / "lib"
    if not lib_base.exists():
        raise FileNotFoundError(f"No 'lib' directory in package {pkg_folder}")

    # Prefer netstandard2.0 if available
    candidates = sorted(lib_base.iterdir())
    for preferred in ["netstandard2.0", "net6.0", "net5.0", "netcoreapp3.1", "net472"]:
        if (lib_base / preferred).exists():
            return lib_base / preferred

    # Fallback: first available folder
    for candidate in candidates:
        if candidate.is_dir():
            return candidate

    raise FileNotFoundError(f"No usable framework folder found in {lib_base}")


def download_and_extract_package(
    package_name: str, version: str, target_dir: Path
) -> Path:
    nupkg_url = f"{NUGET_BASE_URL}/{package_name}/{version}"
    nupkg_path = target_dir / f"{package_name}.{version}.nupkg"

    if not nupkg_path.exists():
        r = requests.get(nupkg_url)
        r.raise_for_status()
        target_dir.mkdir(parents=True, exist_ok=True)
        with open(nupkg_path, "wb") as f:
            f.write(r.content)

    extract_path = target_dir / f"{package_name}_{version}"
    if not extract_path.exists():
        with zipfile.ZipFile(nupkg_path, "r") as zip_ref:
            zip_ref.extractall(extract_path)
    return extract_path


def download_and_load_nuget_package(
    package_name, version, target_dir: Path = None, load_assembly=True
):

    from System.Reflection import Assembly

    if target_dir is None:
        target_dir = nuget_dir

    # Download and extract
    pkg_folder = download_and_extract_package(package_name, version, target_dir)
    lib_folder = find_lib_folder(pkg_folder)

    dll_path = lib_folder / f"{package_name}.dll"
    if not dll_path.exists():
        raise FileNotFoundError(f"{dll_path} not found")

    # sys.path.append(str(lib_folder))
    if load_assembly:
        Assembly.LoadFrom(str(dll_path))  # LoadFile


def is_assembly_loaded(name: str) -> bool:

    import System

    return any(
        asm.GetName().Name == name
        for asm in System.AppDomain.CurrentDomain.GetAssemblies()
    )


_runtime_initialized = False
_nuget_loaded = False


def init_dotnet():
    global _runtime_initialized, _nuget_loaded

    from pythonnet import _LOADED

    # -------------------------------
    # 1. Initialize runtime (ONCE)
    # -------------------------------
    if not _runtime_initialized and not _LOADED:
        from clr_loader import get_coreclr
        from pythonnet import set_runtime

        runtime_config_path = current_dir / "dotnet_lib" / "dotnet.runtime.config.json"
        rt = get_coreclr(runtime_config=str(runtime_config_path))
        set_runtime(rt)

        sempy.fabric._client._utils._init_analysis_services()
        _runtime_initialized = True

    # -------------------------------
    # 2. Load NuGet assemblies (ONCE)
    # -------------------------------
    if _nuget_loaded:
        return

    def load_pkg(name, version):
        if not is_assembly_loaded(name):
            download_and_load_nuget_package(name, version)

    # load_pkg("System.Text.Json", "7.0.4")
    load_pkg("Newtonsoft.Json", "13.0.3")
    load_pkg("NuGet.Versioning", "6.4.0")
    load_pkg("NuGet.Frameworks", "6.4.0")
    load_pkg("System.IO.Packaging", "7.0.0")

    _nuget_loaded = True

    # DAXLib
    DAX_LIB_VERSION = "0.5.3-beta"
    DAX_LIB_ASSEMBLIES = [
        "DaxLib.Client",
    ]

    for name in DAX_LIB_ASSEMBLIES:
        download_and_load_nuget_package(
            name, DAX_LIB_VERSION, nuget_dir, load_assembly=True
        )

    from System.Reflection import Assembly

    # VPAX
    VPA_VERSION = "1.10.0"
    VPA_ASSEMBLIES = [
        "Dax.Metadata",
        "Dax.Model.Extractor",
        "Dax.ViewVpaExport",
        "Dax.Vpax",
    ]

    for name in VPA_ASSEMBLIES:
        download_and_load_nuget_package(
            name, VPA_VERSION, nuget_dir, load_assembly=False
        )

    # For some reason I have to load these after and not inside the download_and_load_nuget_package function
    dll_paths = [
        f"{nuget_dir}/Dax.Model.Extractor_1.10.0/lib/net6.0/Dax.Model.Extractor.dll",
        f"{nuget_dir}/Dax.Metadata_1.10.0/lib/netstandard2.0/Dax.Metadata.dll",
        f"{nuget_dir}/Dax.ViewVpaExport_1.10.0/lib/netstandard2.0/Dax.ViewVpaExport.dll",
        f"{nuget_dir}/Dax.Vpax_1.10.0/lib/net6.0/Dax.Vpax.dll",
    ]
    for dll_path in dll_paths:
        Assembly.LoadFile(dll_path)

    _runtime_set = True



import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _is_valid_uuid,
    _base_api,
    _print_success,
    _create_dataframe,
    resolve_workspace_id,
)
from uuid import UUID
from sempy._utils._log import log


@log
def create_managed_private_endpoint(
    name: str,
    target_private_link_resource_id: UUID,
    target_subresource_type: str,
    request_message: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a managed private endpoint.

    This is a wrapper function for the following API: `Managed Private Endpoints - Create Workspace Managed Private Endpoint <https://learn.microsoft.com/rest/api/fabric/core/managed-private-endpoints/create-workspace-managed-private-endpoint>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str
        Name of the managed private endpoint.
    target_private_link_resource_id: uuid.UUID
        Resource Id of data source for which private endpoint needs to be created.
    target_subresource_type : str
        Sub-resource pointing to Private-link resoure.
    request_message : str, default=None
        Message to approve private endpoint request. Should not be more than 140 characters.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    request_body = {
        "name": name,
        "targetPrivateLinkResourceId": target_private_link_resource_id,
        "targetSubresourceType": target_subresource_type,
    }

    if request_message is not None:
        if len(request_message) > 140:
            raise ValueError(
                f"{icons.red_dot} The request message cannot be more than 140 characters."
            )
        request_body["requestMessage"] = request_message

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/managedPrivateEndpoints",
        method="post",
        status_codes=[201, 202],
        payload=request_body,
        lro_return_status_code=True,
        client="fabric_sp",
    )
    _print_success(
        item_name=name,
        item_type="managed private endpoint",
        workspace_name=workspace_name,
        action="created",
    )


@log
def list_managed_private_endpoints(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Shows the managed private endpoints within a workspace.

    This is a wrapper function for the following API: `Managed Private Endpoints - List Workspace Managed Private Endpoints <https://learn.microsoft.com/rest/api/fabric/core/managed-private-endpoints/list-workspace-managed-private-endpoints>`.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the managed private endpoints within a workspace.
    """

    columns = {
        "Managed Private Endpoint Name": "string",
        "Managed Private Endpoint Id": "string",
        "Target Private Link Resource Id": "string",
        "Provisioning State": "string",
        "Connection Status": "string",
        "Connection Description": "string",
        "Target Subresource Type": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/managedPrivateEndpoints",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            conn = v.get("connectionState", {})
            rows.append(
                {
                    "Managed Private Endpoint Name": v.get("name"),
                    "Managed Private Endpoint Id": v.get("id"),
                    "Target Private Link Resource Id": v.get(
                        "targetPrivateLinkResourceId"
                    ),
                    "Provisioning State": v.get("provisioningState"),
                    "Connection Status": conn.get("status"),
                    "Connection Description": conn.get("description"),
                    "Target Subresource Type": v.get("targetSubresourceType"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def delete_managed_private_endpoint(
    managed_private_endpoint: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes a Fabric managed private endpoint.

    This is a wrapper function for the following API: `Managed Private Endpoints - Delete Workspace Managed Private Endpoint <https://learn.microsoft.com/rest/api/fabric/core/managed-private-endpoints/delete-workspace-managed-private-endpoint>`.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    managed_private_endpoint: str | uuid.UUID
        Name or ID of the managed private endpoint.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    if _is_valid_uuid(managed_private_endpoint):
        item_id = managed_private_endpoint
    else:
        df = list_managed_private_endpoints(workspace=workspace)
        df_filt = df[df["Managed Private Endpoint Name"] == managed_private_endpoint]

        if df_filt.empty:
            raise ValueError(
                f"{icons.red_dot} The '{managed_private_endpoint}' managed private endpoint does not exist within the '{workspace_name}' workspace."
            )

        item_id = df_filt["Managed Private Endpoint Id"].iloc[0]

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/managedPrivateEndpoints/{item_id}",
        method="delete",
        client="fabric_sp",
    )

    _print_success(
        item_name=managed_private_endpoint,
        item_type="managed private endpoint",
        workspace_name=workspace_name,
        action="deleted",
    )


@log
def list_managed_private_endpoint_fqdns(
    managed_private_endpoint: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of fully qualified domain names (FQDNs) associated with the specified managed private endpoint.

    This is a wrapper function for the following API: `Managed Private Endpoints - List FQDNs <https://learn.microsoft.com/rest/api/fabric/core/managed-private-endpoints/list-fqd-ns>`.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    managed_private_endpoint : str | uuid.UUID
        The managed private endpoint name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of fully qualified domain names (FQDNs) associated with the specified managed private endpoint.
    """

    workspace_id = resolve_workspace_id(workspace)
    if _is_valid_uuid(managed_private_endpoint):
        item_id = managed_private_endpoint
    else:
        df = list_managed_private_endpoints(workspace=workspace_id)
        df_filt = df[df["Managed Private Endpoint Name"] == managed_private_endpoint]
        if df_filt.empty:
            raise ValueError(
                f"{icons.red_dot} The '{managed_private_endpoint}' managed private endpoint does not exist within the workspace."
            )
        item_id = df_filt["Managed Private Endpoint Id"].iloc[0]

    columns = {"FQDN": "str"}
    df = _create_dataframe(columns=columns)
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/managedPrivateEndpoints/{item_id}/targetFQDNs",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "FQDN": v,
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



import re
from urllib.parse import urlparse
import sempy.fabric as fabric
from typing import Optional
from uuid import UUID
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    resolve_lakehouse_name_and_id,
    _mount,
    _get_column_aggregate,
    resolve_item_type,
    file_exists,
    create_abfss_path_from_path,
)
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy_labs._a_lib_info import (
    init_dotnet,
)


@log
def create_vpax(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    lakehouse: Optional[str | UUID] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
    file_path: Optional[str] = None,
    read_stats_from_data: bool = False,
    read_direct_query_stats: bool = False,
    direct_lake_stats_mode: str = "ResidentOnly",
    overwrite: bool = False,
):
    """
    Creates a .vpax file for a semantic model and saves it to a lakehouse. This is based on `SQL BI's VertiPaq Analyzer <https://www.sqlbi.com/tools/vertipaq-analyzer/>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str | uuid.UUID, default=None
        The lakehouse name or ID.
        Defaults to None which resolves to the attached lakehouse.
    lakehouse_workspace : str | uuid.UUID, default=None
        The workspace name or ID of the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse.
    file_path : str, default=None
        The path where the .vpax file will be saved in the lakehouse.
        Defaults to None which resolves to the dataset name.
    read_stats_from_data : bool, default=False
        Whether to read statistics from the data.
    read_direct_query_stats : bool, default=False
        Whether to analyze DirectQuery tables.
    direct_lake_stats_mode : str, default='ResidentOnly'
        The Direct Lake extraction mode. Options are 'ResidentOnly' or 'Full'. This parameter is ignored if read_stats_from_data is False. This parameter is only relevant for tables which use Direct Lake mode.
        If set to 'ResidentOnly', column statistics are obtained only for the columns which are in memory.
        If set to 'Full', column statistics are obtained for all columns - pending the proper identification of the Direct Lake source.
    overwrite : bool, default=False
        Whether to overwrite the .vpax file if it already exists in the lakehouse.
    """

    init_dotnet()

    import notebookutils
    from Dax.Metadata import DirectLakeExtractionMode
    from Dax.Model.Extractor import TomExtractor
    from Dax.Vpax.Tools import VpaxTools
    from Dax.ViewVpaExport import Model
    from System.IO import MemoryStream, FileMode, FileStream, FileAccess, FileShare

    direct_lake_stats_mode = direct_lake_stats_mode.capitalize()

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)
    (lakehouse_workspace_name, lakehouse_workspace_id) = resolve_workspace_name_and_id(
        lakehouse_workspace
    )
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=lakehouse_workspace_id
    )

    local_path = _mount(lakehouse=lakehouse_id, workspace=lakehouse_workspace_id)
    if file_path is None:
        file_path = dataset_name

    if file_path.endswith(".vpax"):
        file_path = file_path[:-5]
    save_location = f"Files/{file_path}.vpax"
    path = f"{local_path}/{save_location}"

    # Check if the .vpax file already exists in the lakehouse
    if not overwrite:
        new_path = create_abfss_path_from_path(
            lakehouse_id, lakehouse_workspace_id, save_location
        )
        if file_exists(new_path):
            print(
                f"{icons.warning} The {save_location} file already exists in the '{lakehouse_name}' lakehouse. Set overwrite=True to overwrite the file."
            )
            return

    vpax_stream = MemoryStream()
    extractor_app_name = "VPAX Notebook"
    extractor_app_version = "1.0"
    column_batch_size = 50
    token = notebookutils.credentials.getToken("pbi")
    connection_string = f"data source=powerbi://api.powerbi.com/v1.0/myorg/{workspace_name};initial catalog={dataset_name};User ID=;Password={token};Persist Security Info=True;Impersonation Level=Impersonate"

    print(f"{icons.in_progress} Extracting .vpax metadata...")

    # Get stats for the model; for direct lake only get is_resident
    dax_model = TomExtractor.GetDaxModel(
        connection_string,
        extractor_app_name,
        extractor_app_version,
        read_stats_from_data,
        0,
        read_direct_query_stats,
        DirectLakeExtractionMode.ResidentOnly,
        column_batch_size,
    )
    vpa_model = Model(dax_model)
    tom_database = TomExtractor.GetDatabase(connection_string)

    # Calculate Direct Lake stats for columns which are IsResident=False
    from sempy_labs.tom import connect_semantic_model

    with connect_semantic_model(dataset=dataset, workspace=workspace) as tom:
        is_direct_lake = tom.is_direct_lake()
        if read_stats_from_data and is_direct_lake and direct_lake_stats_mode == "Full":

            df_not_resident = fabric.evaluate_dax(
                dataset=dataset,
                workspace=workspace,
                dax_string=""" SELECT [DIMENSION_NAME] AS [TableName], [ATTRIBUTE_NAME] AS [ColumnName] FROM $SYSTEM.DISCOVER_STORAGE_TABLE_COLUMNS WHERE NOT [ISROWNUMBER] AND NOT [DICTIONARY_ISRESIDENT]""",
            )

            import Microsoft.AnalysisServices.Tabular as TOM

            print(f"{icons.in_progress} Calculating Direct Lake statistics...")

            # For SQL endpoints (do once)
            dfI = fabric.list_items(workspace=workspace)
            # Get list of tables in Direct Lake mode which have columns that are not resident
            tbls = [
                t
                for t in tom.model.Tables
                if t.Name in df_not_resident["TableName"].values
                and any(p.Mode == TOM.ModeType.DirectLake for p in t.Partitions)
            ]
            for t in tbls:
                column_cardinalities = {}
                table_name = t.Name
                partition = next(p for p in t.Partitions)
                entity_name = partition.Source.EntityName
                schema_name = partition.Source.SchemaName
                if len(schema_name) == 0 or schema_name == "dbo":
                    schema_name = None
                expr_name = partition.Source.ExpressionSource.Name
                expr = tom.model.Expressions[expr_name].Expression
                item_id = None
                if "Sql.Database(" in expr:
                    matches = re.findall(r'"([^"]+)"', expr)
                    sql_endpoint_id = matches[1]
                    dfI_filt = dfI[dfI["Id"] == sql_endpoint_id]
                    item_name = (
                        dfI_filt["Display Name"].iloc[0] if not dfI_filt.empty else None
                    )
                    dfI_filt2 = dfI[
                        (dfI["Display Name"] == item_name)
                        & (dfI["Type"].isin(["Lakehouse", "Warehouse"]))
                    ]
                    item_id = dfI_filt2["Id"].iloc[0]
                    item_type = dfI_filt2["Type"].iloc[0]
                    item_workspace_id = workspace_id
                elif "AzureStorage.DataLake(" in expr:
                    match = re.search(r'AzureStorage\.DataLake\("([^"]+)"', expr)
                    if match:
                        url = match.group(1)
                        path_parts = urlparse(url).path.strip("/").split("/")
                        if len(path_parts) >= 2:
                            item_workspace_id, item_id = (
                                path_parts[0],
                                path_parts[1],
                            )
                            item_type = resolve_item_type(
                                item_id=item_id, workspace=workspace_id
                            )
                else:
                    raise NotImplementedError(
                        f"Direct Lake source '{expr}' is not supported. Please report this issue on GitHub (https://github.com/microsoft/semantic-link-labs/issues)."
                    )

                if not item_id:
                    print(
                        f"{icons.info} Cannot determine the Direct Lake source of the '{table_name}' table."
                    )
                elif item_type == "Warehouse":
                    print(
                        f"{icons.info} The '{table_name}' table references a warehouse. Warehouses are not yet supported for this method."
                    )
                else:
                    df_not_resident_cols = df_not_resident[
                        df_not_resident["TableName"] == table_name
                    ]
                    col_dict = {
                        c.Name: c.SourceColumn
                        for c in t.Columns
                        if c.Type != TOM.ColumnType.RowNumber
                        and c.Name in df_not_resident_cols["ColumnName"].values
                    }
                    col_agg = _get_column_aggregate(
                        lakehouse=item_id,
                        workspace=item_workspace_id,
                        table_name=entity_name,
                        schema_name=schema_name,
                        column_name=list(col_dict.values()),
                        function="distinct",
                    )
                    column_cardinalities = {
                        column_name: col_agg[source_column]
                        for column_name, source_column in col_dict.items()
                        if source_column in col_agg
                    }

                    # Update the dax_model file with column cardinalities
                    tbl = next(
                        table
                        for table in dax_model.Tables
                        if str(table.TableName) == table_name
                    )
                    # print(
                    #    f"{icons.in_progress} Calculating column cardinalities for the '{table_name}' table..."
                    # )
                    cols = [
                        col
                        for col in tbl.Columns
                        if str(col.ColumnType) != "RowNumber"
                        and str(col.ColumnName) in column_cardinalities
                    ]
                    for col in cols:
                        # print(str(col.ColumnName), col.ColumnCardinality)
                        col.ColumnCardinality = column_cardinalities.get(
                            str(col.ColumnName)
                        )

    VpaxTools.ExportVpax(vpax_stream, dax_model, vpa_model, tom_database)

    print(f"{icons.in_progress} Exporting .vpax file...")

    mode = FileMode.Create
    file_stream = FileStream(path, mode, FileAccess.Write, FileShare.Read)
    vpax_stream.CopyTo(file_stream)
    file_stream.Close()

    print(
        f"{icons.green_dot} The {file_path}.vpax file has been saved in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace_name}' workspace."
    )


def _dax_distinctcount(table_name, columns):

    dax = "EVALUATE\nROW("
    for c in columns:
        full_name = f"'{table_name}'[{c}]"
        dax += f"""\n"{c}", DISTINCTCOUNT({full_name}),"""

    return f"{dax.rstrip(',')}\n)"



from sempy_labs._helper_functions import (
    _base_api,
    resolve_workspace_id,
    resolve_lakehouse_name_and_id,
    resolve_workspace_name_and_id,
)
from sempy._utils._log import log
from uuid import UUID
from typing import Optional, Literal
import pandas as pd
import sempy_labs._icons as icons


@log
def get_onelake_settings(workspace: Optional[str | UUID] = None):
    """
    Obtains the workspace OneLake settings.

    This is a wrapper function for the following API: `OneLake Settings - Get Settings <https://learn.microsoft.com/rest/api/fabric/core/onelake-settings/get-settings>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        The workspace OneLake settings.
    """

    workspace_id = resolve_workspace_id(workspace)
    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/onelake/settings", client="fabric_sp"
    ).json()

    d = result.get("diagnostics", {})
    enabled = True if d.get("status", {}) == "Enabled" else False
    rows = []
    rows.append(
        {
            "Enabled": enabled,
            "Destination Type": (
                d.get("destination", {}).get("type", {}) if enabled else None
            ),
            "Destination Id": (
                d.get("destination", {}).get("lakehouse", {}).get("itemId", {})
                if enabled
                else None
            ),
            "Destination Workspace Id": (
                d.get("destination", {}).get("lakehouse", {}).get("workspaceId", {})
                if enabled
                else None
            ),
        }
    )

    return pd.DataFrame(rows)


def modify_onelake_diagnostics(
    workspace: Optional[str | UUID] = None,
    enabled: bool = True,
    destination_lakehouse: Optional[str | UUID] = None,
    destination_workspace: Optional[str | UUID] = None,
):
    """
    Obtains the workspace OneLake settings.

    This is a wrapper function for the following API: `OneLake Settings - Modify Diagnostics <https://learn.microsoft.com/rest/api/fabric/core/onelake-settings/modify-diagnostics>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    enabled : bool, default=True
        Whether to enable or disable OneLake diagnostics.
    destination_lakehouse : str | uuid.UUID, default=None
        The name or ID of the destination lakehouse.
        Defaults to None which resolves to the lakehouse of the attached lakehouse
        or if no lakehouse attached, resolves to the lakehouse of the notebook.
    destination_workspace : str | uuid.UUID, default=None
        The name or ID of the destination workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace_id = resolve_workspace_id(workspace)
    (destination_workspace_name, destination_workspace_id) = (
        resolve_workspace_name_and_id(destination_workspace)
    )
    (destination_lakehouse_name, destination_lakehouse_id) = (
        resolve_lakehouse_name_and_id(destination_lakehouse, destination_workspace_id)
    )

    if enabled:
        payload = {
            "status": "Enabled",
            "destination": {
                "type": "Lakehouse",
                "lakehouse": {
                    "referenceType": "ById",
                    "itemId": destination_lakehouse_id,
                    "workspaceId": destination_workspace_id,
                },
            },
        }
    else:
        payload = {"status": "Disabled"}
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/onelake/settings/modifyDiagnostics",
        client="fabric_sp",
        method="post",
        payload=payload,
    )

    if enabled:
        print(
            f"{icons.green_dot} OneLake diagnostics have been enabled and updated to use the '{destination_lakehouse_name}' lakehouse in the '{destination_workspace_name}' workspace as the destination."
        )
    else:
        print(f"{icons.green_dot} OneLake diagnostics have been disabled.")


@log
def modify_immutability_policy(
    retention_days: int,
    scope: Literal["DiagnosticLogs"] = "DiagnosticLogs",
    workspace: Optional[str | UUID] = None,
):
    """
    Create or update OneLake immutability settings.
    Set immutability policy for data stored in OneLake. Currently, this feature supports configuring a retention period specifically for diagnostic logs within a workspace, ensuring they remain unaltered once written.

    This is a wrapper function for the following API: `OneLake Settings - Modify Immutability Policy <https://learn.microsoft.com/rest/api/fabric/core/onelake-settings/modify-immutability-policy>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    retention_days : int
        Retention Days for the action.
    scope : typing.Literal["DiagnosticLogs"], default="DiagnosticLogs"
        The scope of the immutability policy. Currently, only "DiagnosticLogs" is supported.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    payload = {"scope": scope, "retentionDays": retention_days}

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/onelake/settings/modifyImmutabilityPolicy",
        client="fabric_sp",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The OneLake immutability policy has been set to {retention_days} days for scope '{scope}' in the '{workspace_name}' workspace."
    )



import sempy
import sempy.fabric as fabric
import pandas as pd
from typing import List, Optional
from sempy._utils._log import log


@log
def list_all_items(workspaces: Optional[str | List[str]] = None):

    df = pd.DataFrame(
        columns=[
            "Workspace Name",
            "Workspace Id",
            "Item Name",
            "Item Type",
            "Description",
        ]
    )

    if isinstance(workspaces, str):
        workspaces = [workspaces]

    dfW = fabric.list_workspaces()
    if workspaces is not None:
        dfW = dfW[dfW["Name"].isin(workspaces)]

    for _, r in dfW.iterrows():
        workspace_name = r["Name"]
        workspace_id = r["Id"]
        dfI = fabric.list_items(workspace=workspace_name)
        for _, r2 in dfI.iterrows():

            new_data = {
                "Workspace Name": workspace_name,
                "Workspace Id": workspace_id,
                "Item Name": r2["Name"],
                "Item Type": r2["Type"],
                "Description": r2["Description"],
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


@log
def data_dictionary(dataset: str, workspace: Optional[str | None] = None):

    from sempy_labs.tom import connect_semantic_model

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM

    df = pd.DataFrame(
        columns=[
            "Workspace Name",
            "Model Name",
            "Table Name",
            "Object Type",
            "Object Name",
            "Hidden Flag",
            "Description",
            "Display Folder",
            "Measure Formula",
        ]
    )

    with connect_semantic_model(
        dataset=dataset, readonly=True, workspace=workspace
    ) as tom:
        for t in tom.model.Tables:
            expr = None
            if tom.is_calculated_table(table_name=t.Name):
                pName = next(p.Name for p in t.Partitions)
                expr = t.Partitions[pName].Source.Expression

            new_data = {
                "Workspace Name": workspace,
                "Model Name": dataset,
                "Table Name": t.Name,
                "Object Type": t.ObjectType,
                "Object Name": t.Name,
                "Hidden Flag": t.IsHidden,
                "Description": t.Description,
                "Display Folder": None,
                "Measure Formula": expr,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)
            cols = [c for c in t.Columns if c.Type != TOM.ColumnType.RowNumber]
            for c in cols:

                def get_calc_column_expression(table_name, column_name):
                    expr = None
                    if tom.is_calculated_column(
                        table_name=table_name, column_name=column_name
                    ):
                        expr = c.Expression
                    return expr

                new_data = {
                    "Workspace Name": workspace,
                    "Model Name": dataset,
                    "Table Name": t.Name,
                    "Object Type": c.ObjectType,
                    "Object Name": c.Name,
                    "Hidden Flag": c.IsHidden,
                    "Description": c.Description,
                    "Display Folder": c.DisplayFolder,
                    "Measure Formula": get_calc_column_expression(t.Name, c.Name),
                }
                df = pd.concat(
                    [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                )
            for m in t.Measures:
                new_data = {
                    "Workspace Name": workspace,
                    "Model Name": dataset,
                    "Table Name": t.Name,
                    "Object Type": m.ObjectType,
                    "Object Name": m.Name,
                    "Hidden Flag": m.IsHidden,
                    "Description": m.Description,
                    "Display Folder": m.DisplayFolder,
                    "Measure Formula": m.Expression,
                }
                df = pd.concat(
                    [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                )

            if t.CalculationGroup is not None:
                for ci in t.CalculationGroup.CalculationItems:
                    new_data = {
                        "Workspace Name": workspace,
                        "Model Name": dataset,
                        "Table Name": t.Name,
                        "Object Type": "Calculation Item",
                        "Object Name": ci.Name,
                        "Hidden Flag": t.IsHidden,
                        "Description": ci.Description,
                        "Display Folder": None,
                        "Measure Formula": ci.Expression,
                    }
                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

    return df



from ._items import (
    list_graph_models,
    execute_query,
    get_queryable_graph_type,
)
from ._background_jobs import (
    refresh_graph,
)

__all__ = [
    "list_graph_models",
    "execute_query",
    "get_queryable_graph_type",
    "refresh_graph",
]



import pandas as pd
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log


@log
def list_widely_shared_artifacts(
    api_name: str = "LinksSharedToWholeOrganization",
) -> pd.DataFrame:
    """
    Returns a list of Power BI reports that are shared with the whole organization through links or a list of Power BI items (such as reports or dashboards) that are published to the web.

    This is a wrapper function for the following APIs:
    `Admin - WidelySharedArtifacts LinksSharedToWholeOrganization <https://learn.microsoft.com/rest/api/power-bi/admin/widely-shared-artifacts-links-shared-to-whole-organization>`_.
    `Admin - WidelySharedArtifacts PublishedToWeb <https://learn.microsoft.com/rest/api/power-bi/admin/widely-shared-artifacts-published-to-web>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    api_name : str, default = "LinksSharedToWholeOrganization"
        The name of the API to call. Either "LinksSharedToWholeOrganization" or "PublishedToWeb".

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of Power BI reports that are shared with the whole organization through links or a list of Power BI items (such as reports or dashboards) that are published to the web.
    """

    columns = {
        "Artifact Id": "string",
        "Artifact Name": "string",
        "Artifact Type": "string",
        "Access Right": "string",
        "Share Type": "string",
        "Sharer Name": "string",
        "Sharer Email Address": "string",
        "Sharer Identifier": "string",
        "Sharer Graph Id": "string",
        "Sharer Principal Type": "string",
    }

    df = _create_dataframe(columns=columns)

    api = (
        "linksSharedToWholeOrganization"
        if api_name == "LinksSharedToWholeOrganization"
        else "publishedToWeb"
    )

    responses = _base_api(
        request=f"/v1.0/myorg/admin/widelySharedArtifacts/{api}",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("ArtifactAccessEntities", []):
            sharer = v.get("sharer", {})
            rows.append(
                {
                    "Artifact Id": v.get("artifactId"),
                    "Artifact Name": v.get("displayName"),
                    "Artifact Type": v.get("artifactType"),
                    "Access Right": v.get("accessRight"),
                    "Share Type": v.get("shareType"),
                    "Sharer Name": sharer.get("displayName"),
                    "Sharer Email Address": sharer.get("emailAddress"),
                    "Sharer Identifier": sharer.get("identifier"),
                    "Sharer Graph Id": sharer.get("graphId"),
                    "Sharer Principal Type": sharer.get("principalType"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



from sempy_labs._helper_functions import (
    resolve_item_id,
    _base_api,
    resolve_workspace_id,
)
from typing import Optional, Literal
import sempy_labs._icons as icons
from uuid import UUID
from sempy._utils._log import log


@log
def get_connection_string(
    item: str | UUID,
    type: Literal["Lakehouse", "Warehouse", "SQLEndpoint"],
    workspace: Optional[str | UUID] = None,
    guest_tenant_id: Optional[UUID] = None,
    private_link_type: Optional[str] = None,
) -> str:
    """
    Returns the SQL connection string of the specified item.

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item (Lakehouse or MirroredDatabase).
    type : typing.Literal['Lakehouse', 'Warehouse', 'SQLEndpoint']
        The type of the item. Must be 'Lakehouse' or 'MirroredDatabase'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    guest_tenant_id : uuid.UUID, default=None
        The guest tenant ID if the end user's tenant is different from the warehouse tenant.
    private_link_type : str, default=None
        Indicates the type of private link this connection string uses. Must be either 'Workspace' or 'None' or left as None.

    Returns
    -------
    str
        Returns the SQL connection string of the specified item.
    """
    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=item, type=type, workspace=workspace)

    type_dict = {
        "Warehouse": "warehouses",
        "SQLEndpoint": "sqlEndpoints",
    }
    type_for_url = type_dict.get(type)

    if type == "Lakehouse":
        response = _base_api(
            f"/v1/workspaces/{workspace_id}/lakehouses/{item_id}", client="fabric_sp"
        ).json()
        return (
            response.get("properties", {})
            .get("sqlEndpointProperties", {})
            .get("connectionString", {})
        )
    if type in ["SQLEndpoint", "Warehouse"]:
        url = f"/v1/workspaces/{workspace_id}/{type_for_url}/{item_id}/connectionString"
    else:
        raise ValueError(
            f"{icons.red_dot} The type must be 'Lakehouse', 'Warehouse' or 'SQLEndpoint'."
        )

    if private_link_type is not None and private_link_type not in ["Workspace", "None"]:
        raise ValueError(
            f"{icons.red_dot} private_link_type must be 'Workspace' or 'None' or left as None."
        )

    if guest_tenant_id or private_link_type:
        params = []
        if guest_tenant_id:
            params.append(f"guestTenantId={guest_tenant_id}")
        if private_link_type:
            params.append(f"privateLinkType={private_link_type}")
        param_str = "?" + "&".join(params)
        url += param_str

    response = _base_api(request=url, client="fabric_sp")

    return response.json().get("connectionString")



from ._items import (
    list_apache_airflow_jobs,
    delete_apache_airflow_job,
)
from ._files import (
    list_apache_airflow_job_files,
    delete_apache_airflow_job_file,
    create_or_update_apache_airflow_job_file,
)

__all__ = [
    "list_apache_airflow_jobs",
    "delete_apache_airflow_job",
    "list_apache_airflow_job_files",
    "delete_apache_airflow_job_file",
    "create_or_update_apache_airflow_job_file",
]



from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    resolve_lakehouse_name_and_id,
    resolve_item_name_and_id,
    _get_fabric_context_setting,
)
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
from sempy._utils._log import log


@log
def generate_shared_expression(
    item_name: Optional[str] = None,
    item_type: str = "Lakehouse",
    workspace: Optional[str | UUID] = None,
    use_sql_endpoint: bool = True,
) -> str:
    """
    Dynamically generates the M expression used by a Direct Lake model for a given lakehouse/warehouse.

    Parameters
    ----------
    item_name : str, default=None
        The Fabric lakehouse or warehouse name.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    item_type : str, default="Lakehouse"
        The Fabric item name. Valid options: 'Lakehouse', 'Warehouse'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the item.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    use_sql_endpoint : bool, default=True
        Whether to use the SQL Endpoint for the lakehouse/warehouse.
        If False, the expression will be generated without using the SQL Endpoint.

    Returns
    -------
    str
        Shows the expression which can be used to connect a Direct Lake semantic model to its SQL Endpoint.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_types = ["Lakehouse", "Warehouse"]
    item_type = item_type.capitalize()
    if item_type not in item_types:
        raise ValueError(
            f"{icons.red_dot} Invalid item type. Valid options: {item_types}."
        )

    if item_type == "Lakehouse":
        (item_name, item_id) = resolve_lakehouse_name_and_id(
            lakehouse=item_name, workspace=workspace_id
        )
    else:
        (item_name, item_id) = resolve_item_name_and_id(
            item=item_name, type=item_type, workspace=workspace_id
        )

    if use_sql_endpoint:
        item_type_rest = f"{item_type.lower()}s"
        response = _base_api(
            request=f"/v1/workspaces/{workspace_id}/{item_type_rest}/{item_id}"
        )

        prop = response.json().get("properties")

        if item_type == "Lakehouse":
            sqlprop = prop.get("sqlEndpointProperties")
            sqlEPCS = sqlprop.get("connectionString")
            sqlepid = sqlprop.get("id")
            provStatus = sqlprop.get("provisioningStatus")
        elif item_type == "Warehouse":
            sqlEPCS = prop.get("connectionString")
            sqlepid = item_id
            provStatus = None

        if provStatus == "InProgress":
            raise ValueError(
                f"{icons.red_dot} The SQL Endpoint for the '{item_name}' {item_type.lower()} within the '{workspace_name}' workspace has not yet been provisioned. Please wait until it has been provisioned."
            )

        start_expr = "let\n\tdatabase = "
        end_expr = "\nin\n\tdatabase"
        mid_expr = f'Sql.Database("{sqlEPCS}", "{sqlepid}")'
        return f"{start_expr}{mid_expr}{end_expr}"
    else:
        # Build DL/OL expression
        env = _get_fabric_context_setting("spark.trident.pbienv").lower()
        env = "" if env == "prod" else f"{env}-"

        return f"""let\n\tSource = AzureStorage.DataLake("https://{env}onelake.dfs.fabric.microsoft.com/{workspace_id}/{item_id}")\nin\n\tSource"""



from ._restore_points import (
    create_restore_point,
    delete_restore_point,
    list_restore_points,
    update_restore_point,
    restore_to_restore_point,
)
from ._items import (
    create_warehouse,
    delete_warehouse,
    get_warehouse_tables,
    get_warehouse_columns,
    list_warehouses,
)


__all__ = [
    "create_restore_point",
    "delete_restore_point",
    "list_restore_points",
    "update_restore_point",
    "restore_to_restore_point",
    "create_warehouse",
    "delete_warehouse",
    "get_warehouse_tables",
    "get_warehouse_columns",
    "list_warehouses",
]



import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs.spark as sp


@log
def list_custom_pools(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Lists all `custom pools <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - List Workspace Custom Pools <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/list-workspace-custom-pools>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all the custom pools within the Fabric workspace.
    """

    return sp.list_custom_pools(workspace=workspace)


@log
def create_custom_pool(
    pool_name: str,
    node_size: str,
    min_node_count: int,
    max_node_count: int,
    min_executors: int,
    max_executors: int,
    node_family: str = "MemoryOptimized",
    auto_scale_enabled: bool = True,
    dynamic_executor_allocation_enabled: bool = True,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a `custom pool <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - Create Workspace Custom Pool <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool>`_.

    Parameters
    ----------
    pool_name : str
        The custom pool name.
    node_size : str
        The `node size <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodesize>`_.
    min_node_count : int
        The `minimum node count <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
    max_node_count : int
        The maximum node count.
    min_executors : int
        The `minimum executors <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
    max_executors : int
        The maximum executors.
    node_family : str, default='MemoryOptimized'
        The `node family <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodefamily>`_.
    auto_scale_enabled : bool, default=True
        The status of `auto scale <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
    dynamic_executor_allocation_enabled : bool, default=True
        The status of the `dynamic executor allocation <https://learn.microsoft.com/en-us/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sp.create_custom_pool(
        pool_name=pool_name,
        node_size=node_size,
        min_node_count=min_node_count,
        max_node_count=max_node_count,
        min_executors=min_executors,
        max_executors=max_executors,
        node_family=node_family,
        auto_scale_enabled=auto_scale_enabled,
        dynamic_executor_allocation_enabled=dynamic_executor_allocation_enabled,
        workspace=workspace,
    )


@log
def update_custom_pool(
    pool_name: str,
    node_size: Optional[str] = None,
    min_node_count: Optional[int] = None,
    max_node_count: Optional[int] = None,
    min_executors: Optional[int] = None,
    max_executors: Optional[int] = None,
    node_family: Optional[str] = None,
    auto_scale_enabled: Optional[bool] = None,
    dynamic_executor_allocation_enabled: Optional[bool] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the properties of a `custom pool <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - Update Workspace Custom Pool <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/update-workspace-custom-pool>`_.

    Parameters
    ----------
    pool_name : str
        The custom pool name.
    node_size : str, default=None
        The `node size <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodesize>`_.
        Defaults to None which keeps the existing property setting.
    min_node_count : int, default=None
        The `minimum node count <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
        Defaults to None which keeps the existing property setting.
    max_node_count : int, default=None
        The maximum node count.
        Defaults to None which keeps the existing property setting.
    min_executors : int, default=None
        The `minimum executors <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
        Defaults to None which keeps the existing property setting.
    max_executors : int, default=None
        The maximum executors.
        Defaults to None which keeps the existing property setting.
    node_family : str, default=None
        The `node family <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodefamily>`_.
        Defaults to None which keeps the existing property setting.
    auto_scale_enabled : bool, default=None
        The status of `auto scale <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
        Defaults to None which keeps the existing property setting.
    dynamic_executor_allocation_enabled : bool, default=None
        The status of the `dynamic executor allocation <https://learn.microsoft.com/en-us/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
        Defaults to None which keeps the existing property setting.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sp.update_custom_pool(
        pool_name=pool_name,
        node_size=node_size,
        min_node_count=min_node_count,
        max_node_count=max_node_count,
        min_executors=min_executors,
        max_executors=max_executors,
        node_family=node_family,
        auto_scale_enabled=auto_scale_enabled,
        dynamic_executor_allocation_enabled=dynamic_executor_allocation_enabled,
        workspace=workspace,
    )


@log
def delete_custom_pool(pool_name: str, workspace: Optional[str | UUID] = None):
    """
    Deletes a `custom pool <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - Delete Workspace Custom Pool <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/delete-workspace-custom-pool>`_.

    Parameters
    ----------
    pool_name : str
        The custom pool name.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sp.delete_custom_pool(pool_name=pool_name, workspace=workspace)


@log
def get_spark_settings(
    workspace: Optional[str | UUID] = None, return_dataframe: bool = True
) -> pd.DataFrame | dict:
    """
    Shows the spark settings for a workspace.

    This is a wrapper function for the following API: `Workspace Settings - Get Spark Settings <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/get-spark-settings>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    return_dataframe : bool, default=True
        If True, returns a pandas dataframe. If False, returns a json dictionary.

    Returns
    -------
    pandas.DataFrame | dict
        A pandas dataframe showing the spark settings for a workspace.
    """

    return sp.get_spark_settings(workspace=workspace, return_dataframe=return_dataframe)


@log
def update_spark_settings(
    automatic_log_enabled: Optional[bool] = None,
    high_concurrency_enabled: Optional[bool] = None,
    customize_compute_enabled: Optional[bool] = None,
    default_pool_name: Optional[str] = None,
    max_node_count: Optional[int] = None,
    max_executors: Optional[int] = None,
    environment_name: Optional[str] = None,
    runtime_version: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the spark settings for a workspace.

    This is a wrapper function for the following API: `Workspace Settings - Update Spark Settings <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings>`_.

    Parameters
    ----------
    automatic_log_enabled : bool, default=None
        The status of the `automatic log <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#automaticlogproperties>`_.
        Defaults to None which keeps the existing property setting.
    high_concurrency_enabled : bool, default=None
        The status of the `high concurrency <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#highconcurrencyproperties>`_ for notebook interactive run.
        Defaults to None which keeps the existing property setting.
    customize_compute_enabled : bool, default=None
        `Customize compute <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#poolproperties>`_ configurations for items.
        Defaults to None which keeps the existing property setting.
    default_pool_name : str, default=None
        `Default pool <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#poolproperties>`_ for workspace.
        Defaults to None which keeps the existing property setting.
    max_node_count : int, default=None
        The maximum node count.
        Defaults to None which keeps the existing property setting.
    max_executors : int, default=None
        The maximum executors.
        Defaults to None which keeps the existing property setting.
    environment_name : str, default=None
        The name of the `default environment <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#environmentproperties>`_. Empty string indicated there is no workspace default environment
        Defaults to None which keeps the existing property setting.
    runtime_version : str, default=None
        The `runtime version <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#environmentproperties>`_.
        Defaults to None which keeps the existing property setting.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sp.update_spark_settings(
        automatic_log_enabled=automatic_log_enabled,
        high_concurrency_enabled=high_concurrency_enabled,
        customize_compute_enabled=customize_compute_enabled,
        default_pool_name=default_pool_name,
        max_node_count=max_node_count,
        max_executors=max_executors,
        environment_name=environment_name,
        runtime_version=runtime_version,
        workspace=workspace,
    )



import pandas as pd
from typing import Any, Optional, List
from sempy_labs._helper_functions import (
    _update_dataframe_datatypes,
    resolve_item_id,
    resolve_item_name_and_id,
    resolve_workspace_id,
    _base_api,
    delete_item,
    _create_dataframe,
    create_item,
    resolve_workspace_name_and_id,
)
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def list_ml_models(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the ML models within a workspace.

    This is a wrapper function for the following API: `Items - List ML Models <https://learn.microsoft.com/rest/api/fabric/mlmodel/items/list-ml-models>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the ML models within a workspace.
    """

    columns = {
        "ML Model Name": "string",
        "ML Model Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlModels",
        status_codes=200,
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            model_id = v.get("id")
            modelName = v.get("displayName")
            desc = v.get("description")

            rows.append(
                {
                    "ML Model Name": modelName,
                    "ML Model Id": model_id,
                    "Description": desc,
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def create_ml_model(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric ML model.

    This is a wrapper function for the following API: `Items - Create ML Model <https://learn.microsoft.com/rest/api/fabric/mlmodel/items/create-ml-model>`_.

    Parameters
    ----------
    name: str
        Name of the ML model.
    description : str, default=None
        A description of the ML model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(name=name, description=description, type="MLModel", workspace=workspace)


@log
def delete_ml_model(ml_model: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a Fabric ML model.

    This is a wrapper function for the following API: `Items - Delete ML Model <https://learn.microsoft.com/rest/api/fabric/mlmodel/items/delete-ml-model>`_.

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=ml_model, type="MLModel", workspace=workspace)


@log
def activate_ml_model_endpoint_version(
    ml_model: str | UUID, name: str, workspace: Optional[str | UUID] = None
):
    """
    Activates the specified model version endpoint.

    This is a wrapper function for the following API: `Endpoint - Activate ML Model Endpoint Version <https://learn.microsoft.com/rest/api/fabric/mlmodel/endpoint/activate-ml-model-endpoint-version>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    name: str
        The ML model version name.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (model_name, model_id) = resolve_item_name_and_id(
        item=ml_model, type="MLModel", workspace=workspace
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlmodels/{model_id}/endpoint/versions/{name}/activate",
        method="post",
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} The {model_name} model version {name} has been activated in the {workspace_name} workspace."
    )


@log
def deactivate_ml_model_endpoint_version(
    ml_model: str | UUID, name: str, workspace: Optional[str | UUID] = None
):
    """
    Deactivates the specified model version endpoint.

    This is a wrapper function for the following API: `Endpoint - Deactivate ML Model Endpoint Version <https://learn.microsoft.com/rest/api/fabric/mlmodel/endpoint/deactivate-ml-model-endpoint-version>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    name: str
        The ML model version name.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (model_name, model_id) = resolve_item_name_and_id(
        item=ml_model, type="MLModel", workspace=workspace
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlmodels/{model_id}/endpoint/versions/{name}/deactivate",
        method="post",
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} The {model_name} model version {name} has been deactivated in the {workspace_name} workspace."
    )


@log
def deactivate_all_ml_model_endpoint_versions(
    ml_model: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deactivates the specified machine learning model and its version's endpoints.

    This is a wrapper function for the following API: `Endpoint - Deactivate All ML Model Endpoint Versions <https://learn.microsoft.com/rest/api/fabric/mlmodel/endpoint/deactivate-all-ml-model-endpoint-versions>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (model_name, model_id) = resolve_item_name_and_id(
        item=ml_model, type="MLModel", workspace=workspace
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlmodels/{model_id}/endpoint/versions/deactivateAll",
        method="post",
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} All endpoint versions of the {model_name} model within the {workspace_name} workspace have been deactivated."
    )


@log
def list_ml_model_endpoint_versions(
    ml_model: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Lists all machine learning model endpoint versions.

    This is a wrapper function for the following API: `Endpoint - List ML Model Endpoint Versions <https://learn.microsoft.com/rest/api/fabric/mlmodel/endpoint/list-ml-model-endpoint-versions>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the ML model endpoint versions within a workspace.
    """

    workspace_id = resolve_workspace_id(workspace)
    model_id = resolve_item_id(item=ml_model, type="MLModel", workspace=workspace)

    columns = {
        "Version Name": "string",
        "Status": "string",
        "Type": "string",
        "Name": "string",
        "Required": "bool",
        "Scale Rule": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlmodels/{model_id}/endpoint/versions",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for version in r.get("value", []):
            base = {
                "Version Name": version.get("versionName"),
                "Status": version.get("status"),
                "Scale Rule": version.get("scaleRule"),
            }
            for sig_type in ["inputSignature", "outputSignature"]:
                for entry in version.get(sig_type, []):
                    rows.append(
                        {
                            **base,
                            "Signature Type": (
                                "Input" if sig_type == "inputSignature" else "Output"
                            ),
                            "Name": entry.get("name"),
                            "Type": entry.get("type"),
                            "Required": entry.get("required"),
                        }
                    )
            # Handle versions with no signatures
            if "inputSignature" not in version and "outputSignature" not in version:
                rows.append(base)

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def score_ml_model_endpoint(
    ml_model: str | UUID,
    inputs: List[List[Any]],
    orientation: str = "values",
    workspace: Optional[str | UUID] = None,
) -> dict:
    """
    Scores input data using the default version of the endpoint and returns results.

    This is a wrapper function for the following API: `Endpoint - Score ML Model Endpoint <https://learn.microsoft.com/rest/api/fabric/mlmodel/endpoint/score-ml-model-endpoint>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    inputs: List[List[Any]]
        Machine learning inputs to score in the form of Pandas dataset arrays that can include strings, numbers, integers and booleans.
    orientation: str, default='values'
        `Orientation <https://learn.microsoft.com/en-us/rest/api/fabric/mlmodel/endpoint/score-ml-model-endpoint?tabs=HTTP#orientation>`_ of the input data.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace_id = resolve_workspace_id(workspace)
    model_id = resolve_item_id(item=ml_model, type="MLModel", workspace=workspace)

    orientation = _validate_orientation(orientation)
    payload = {
        "formatType": "dataframe",
        "orientation": orientation,
        "inputs": inputs,
    }

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlmodels/{model_id}/endpoint/score",
        method="post",
        client="fabric_sp",
        payload=payload,
        lro_return_json=True,
        status_codes=[200, 202],
    )

    return result


@log
def score_ml_model_endpoint_version(
    ml_model: str | UUID,
    name: str,
    inputs: List[List[Any]],
    orientation: str = "values",
    workspace: Optional[str | UUID] = None,
) -> dict:
    """
    Scores input data using the default version of the endpoint and returns results.

    This is a wrapper function for the following API: `Endpoint - Score ML Model Endpoint Version <https://learn.microsoft.com/rest/api/fabric/mlmodel/endpoint/score-ml-model-endpoint-version>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    ml_model: str | uuid.UUID
        Name or ID of the ML model.
    name: str
        The ML model version name.
    inputs: List[List[Any]]
        Machine learning inputs to score in the form of Pandas dataset arrays that can include strings, numbers, integers and booleans.
    orientation: str, default='values'
        `Orientation <https://learn.microsoft.com/en-us/rest/api/fabric/mlmodel/endpoint/score-ml-model-endpoint?tabs=HTTP#orientation>`_ of the input data.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace_id = resolve_workspace_id(workspace)
    model_id = resolve_item_id(item=ml_model, type="MLModel", workspace=workspace)

    orientation = _validate_orientation(orientation)
    payload = {
        "formatType": "dataframe",
        "orientation": orientation,
        "inputs": inputs,
    }

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlmodels/{model_id}/endpoint/versions/{name}/score",
        method="post",
        client="fabric_sp",
        payload=payload,
        lro_return_json=True,
        status_codes=[200, 202],
    )

    return result


def _validate_orientation(orientation: str):

    orientation = orientation.lower()
    if orientation not in ["split", "values", "record", "index", "table"]:
        raise ValueError(
            f"Invalid orientation '{orientation}'. Must be one of 'split', 'values', 'record', 'index', or 'table'."
        )
    return orientation
